@@ -42,6 +42,6 @@ async fn main() -> Result<()> {
     Toplevel::new()
         .start("Countdown", |h| CountdownSubsystem::new().run(h))
         .catch_signals()
-        .wait_for_shutdown(Duration::from_millis(1000))
+        .handle_shutdown_requests(Duration::from_millis(1000))
         .await
 }
