@@ -1,12 +1,27 @@
 use anyhow::anyhow;
+use env_logger;
 use tokio::time::{sleep, timeout, Duration};
-use tokio_graceful_shutdown::{SubsystemHandle, Toplevel};
+use tokio_graceful_shutdown::{PartialShutdownError, SubsystemHandle, Toplevel};
 
 mod common;
 use common::event::Event;
 
+use std::sync::Once;
+
+static INIT: Once = Once::new();
+
+/// Setup function that is only run once, even if called multiple times.
+fn setup() {
+    INIT.call_once(|| {
+        // Init logging
+        env_logger::Builder::from_env(env_logger::Env::default().default_filter_or("off")).init();
+    });
+}
+
 #[tokio::test]
 async fn normal_shutdown() {
+    setup();
+
     let subsystem = |subsys: SubsystemHandle| async move {
         subsys.on_shutdown_requested().await;
         sleep(Duration::from_millis(200)).await;
@@ -22,7 +37,9 @@ async fn normal_shutdown() {
             shutdown_token.shutdown();
         },
         async {
-            let result = toplevel.wait_for_shutdown(Duration::from_millis(400)).await;
+            let result = toplevel
+                .handle_shutdown_requests(Duration::from_millis(400))
+                .await;
             assert!(result.is_ok());
         },
     );
@@ -30,6 +47,8 @@ async fn normal_shutdown() {
 
 #[tokio::test]
 async fn shutdown_timeout_causes_error() {
+    setup();
+
     let subsystem = |subsys: SubsystemHandle| async move {
         subsys.on_shutdown_requested().await;
         sleep(Duration::from_millis(400)).await;
@@ -45,7 +64,9 @@ async fn shutdown_timeout_causes_error() {
             shutdown_token.shutdown();
         },
         async {
-            let result = toplevel.wait_for_shutdown(Duration::from_millis(200)).await;
+            let result = toplevel
+                .handle_shutdown_requests(Duration::from_millis(200))
+                .await;
             assert!(result.is_err());
         },
     );
@@ -53,6 +74,8 @@ async fn shutdown_timeout_causes_error() {
 
 #[tokio::test]
 async fn subsystem_finishes_with_success() {
+    setup();
+
     let subsystem = |_| async { Ok(()) };
 
     let (toplevel_finished, set_toplevel_finished) = Event::create();
@@ -62,7 +85,9 @@ async fn subsystem_finishes_with_success() {
 
     tokio::join!(
         async {
-            let result = toplevel.wait_for_shutdown(Duration::from_millis(100)).await;
+            let result = toplevel
+                .handle_shutdown_requests(Duration::from_millis(100))
+                .await;
             set_toplevel_finished();
             // Assert Ok(()) returncode properly propagates to Toplevel
             assert!(result.is_ok());
@@ -81,6 +106,8 @@ async fn subsystem_finishes_with_success() {
 
 #[tokio::test]
 async fn subsystem_finishes_with_error() {
+    setup();
+
     let subsystem = |_| async { Err(anyhow!("Error!")) };
 
     let (toplevel_finished, set_toplevel_finished) = Event::create();
@@ -90,7 +117,9 @@ async fn subsystem_finishes_with_error() {
 
     tokio::join!(
         async {
-            let result = toplevel.wait_for_shutdown(Duration::from_millis(100)).await;
+            let result = toplevel
+                .handle_shutdown_requests(Duration::from_millis(100))
+                .await;
             set_toplevel_finished();
             // Assert Err(()) returncode properly propagates to Toplevel
             assert!(result.is_err());
@@ -106,6 +135,8 @@ async fn subsystem_finishes_with_error() {
 
 #[tokio::test]
 async fn subsystem_receives_shutdown() {
+    setup();
+
     let (subsys_finished, set_subsys_finished) = Event::create();
 
     let subsys = |subsys: SubsystemHandle| async move {
@@ -116,7 +147,7 @@ async fn subsystem_receives_shutdown() {
 
     let toplevel = Toplevel::new().start("subsys", subsys);
     let shutdown_token = toplevel.get_shutdown_token().clone();
-    let result = tokio::spawn(toplevel.wait_for_shutdown(Duration::from_millis(100)));
+    let result = tokio::spawn(toplevel.handle_shutdown_requests(Duration::from_millis(100)));
 
     sleep(Duration::from_millis(100)).await;
     assert!(!subsys_finished.get());
@@ -136,6 +167,8 @@ async fn subsystem_receives_shutdown() {
 
 #[tokio::test]
 async fn nested_subsystem_receives_shutdown() {
+    setup();
+
     let (subsys_finished, set_subsys_finished) = Event::create();
 
     let nested_subsystem = |subsys: SubsystemHandle| async move {
@@ -152,7 +185,7 @@ async fn nested_subsystem_receives_shutdown() {
 
     let toplevel = Toplevel::new().start("subsys", subsystem);
     let shutdown_token = toplevel.get_shutdown_token().clone();
-    let result = tokio::spawn(toplevel.wait_for_shutdown(Duration::from_millis(100)));
+    let result = tokio::spawn(toplevel.handle_shutdown_requests(Duration::from_millis(100)));
 
     sleep(Duration::from_millis(100)).await;
     assert!(!subsys_finished.get());
@@ -172,6 +205,8 @@ async fn nested_subsystem_receives_shutdown() {
 
 #[tokio::test]
 async fn nested_subsystem_error_propagates() {
+    setup();
+
     let nested_subsystem = |_subsys: SubsystemHandle| async move { Err(anyhow!("Error!")) };
 
     let subsystem = move |mut subsys: SubsystemHandle| async move {
@@ -187,7 +222,9 @@ async fn nested_subsystem_error_propagates() {
 
     tokio::join!(
         async {
-            let result = toplevel.wait_for_shutdown(Duration::from_millis(100)).await;
+            let result = toplevel
+                .handle_shutdown_requests(Duration::from_millis(100))
+                .await;
             set_toplevel_finished();
             // Assert Err(()) returncode properly propagates to Toplevel
             assert!(result.is_err());
@@ -203,6 +240,8 @@ async fn nested_subsystem_error_propagates() {
 
 #[tokio::test]
 async fn panic_gets_handled_correctly() {
+    setup();
+
     let nested_subsystem = |_subsys: SubsystemHandle| async move {
         panic!("Error!");
     };
@@ -220,7 +259,9 @@ async fn panic_gets_handled_correctly() {
 
     tokio::join!(
         async {
-            let result = toplevel.wait_for_shutdown(Duration::from_millis(100)).await;
+            let result = toplevel
+                .handle_shutdown_requests(Duration::from_millis(100))
+                .await;
             set_toplevel_finished();
             // Assert panic causes Error propagation to Toplevel
             assert!(result.is_err());
@@ -236,6 +277,8 @@ async fn panic_gets_handled_correctly() {
 
 #[tokio::test]
 async fn subsystem_can_request_shutdown() {
+    setup();
+
     let (subsystem_should_stop, stop_subsystem) = Event::create();
 
     let (subsys_finished, set_subsys_finished) = Event::create();
@@ -255,7 +298,9 @@ async fn subsystem_can_request_shutdown() {
 
     tokio::join!(
         async {
-            let result = toplevel.wait_for_shutdown(Duration::from_millis(100)).await;
+            let result = toplevel
+                .handle_shutdown_requests(Duration::from_millis(100))
+                .await;
             set_toplevel_finished();
 
             // Assert graceful shutdown does not cause an Error code
@@ -280,6 +325,8 @@ async fn subsystem_can_request_shutdown() {
 
 #[tokio::test]
 async fn shutdown_timeout_causes_cancellation() {
+    setup();
+
     let (subsys_finished, set_subsys_finished) = Event::create();
 
     let subsystem = |subsys: SubsystemHandle| async move {
@@ -296,7 +343,9 @@ async fn shutdown_timeout_causes_cancellation() {
 
     tokio::join!(
         async {
-            let result = toplevel.wait_for_shutdown(Duration::from_millis(200)).await;
+            let result = toplevel
+                .handle_shutdown_requests(Duration::from_millis(200))
+                .await;
             set_toplevel_finished();
 
             // Assert graceful shutdown does not cause an Error code
@@ -326,6 +375,8 @@ async fn shutdown_timeout_causes_cancellation() {
 
 #[tokio::test]
 async fn spawning_task_during_shutdown_causes_task_to_be_cancelled() {
+    setup();
+
     let (subsys_finished, set_subsys_finished) = Event::create();
     let (nested_finished, set_nested_finished) = Event::create();
 
@@ -350,7 +401,9 @@ async fn spawning_task_during_shutdown_causes_task_to_be_cancelled() {
 
     tokio::join!(
         async {
-            let result = toplevel.wait_for_shutdown(Duration::from_millis(500)).await;
+            let result = toplevel
+                .handle_shutdown_requests(Duration::from_millis(500))
+                .await;
             set_toplevel_finished();
 
             // Assert graceful shutdown does not cause an Error code
@@ -382,6 +435,8 @@ async fn spawning_task_during_shutdown_causes_task_to_be_cancelled() {
 
 #[tokio::test(flavor = "multi_thread", worker_threads = 3)]
 async fn double_panic_does_not_stop_graceful_shutdown() {
+    setup();
+
     let (subsys_finished, set_subsys_finished) = Event::create();
 
     let subsys3 = |subsys: SubsystemHandle| async move {
@@ -406,7 +461,7 @@ async fn double_panic_does_not_stop_graceful_shutdown() {
 
     let result = Toplevel::new()
         .start("subsys", subsys1)
-        .wait_for_shutdown(Duration::from_millis(500))
+        .handle_shutdown_requests(Duration::from_millis(500))
         .await;
     assert!(result.is_err());
 
@@ -415,6 +470,8 @@ async fn double_panic_does_not_stop_graceful_shutdown() {
 
 #[tokio::test]
 async fn destroying_toplevel_cancels_subsystems() {
+    setup();
+
     let (subsys_started, set_subsys_started) = Event::create();
     let (subsys_finished, set_subsys_finished) = Event::create();
 
@@ -436,6 +493,8 @@ async fn destroying_toplevel_cancels_subsystems() {
 
 #[tokio::test]
 async fn destroying_toplevel_cancels_nested_toplevel_subsystems() {
+    setup();
+
     let (subsys_started, set_subsys_started) = Event::create();
     let (subsys_finished, set_subsys_finished) = Event::create();
 
@@ -449,7 +508,7 @@ async fn destroying_toplevel_cancels_nested_toplevel_subsystems() {
     let subsys1 = move |_subsys: SubsystemHandle| async move {
         Toplevel::new()
             .start("subsys2", subsys2)
-            .wait_for_shutdown(Duration::from_millis(100))
+            .handle_shutdown_requests(Duration::from_millis(100))
             .await
     };
 
@@ -461,3 +520,260 @@ async fn destroying_toplevel_cancels_nested_toplevel_subsystems() {
     assert!(subsys_started.get());
     assert!(!subsys_finished.get());
 }
+
+#[tokio::test]
+async fn partial_shutdown_request_stops_nested_subsystems() {
+    setup();
+
+    let (subsys1_started, set_subsys1_started) = Event::create();
+    let (subsys1_finished, set_subsys1_finished) = Event::create();
+    let (subsys2_started, set_subsys2_started) = Event::create();
+    let (subsys2_finished, set_subsys2_finished) = Event::create();
+    let (subsys3_started, set_subsys3_started) = Event::create();
+    let (subsys3_finished, set_subsys3_finished) = Event::create();
+    let (subsys1_shutdown_performed, set_subsys1_shutdown_performed) = Event::create();
+
+    let subsys3 = move |subsys: SubsystemHandle| async move {
+        set_subsys3_started();
+        subsys.on_shutdown_requested().await;
+        set_subsys3_finished();
+        Ok(())
+    };
+    let subsys2 = move |mut subsys: SubsystemHandle| async move {
+        set_subsys2_started();
+        subsys.start("subsys3", subsys3);
+        subsys.on_shutdown_requested().await;
+        set_subsys2_finished();
+        Ok(())
+    };
+
+    let subsys1 = move |mut subsys: SubsystemHandle| async move {
+        set_subsys1_started();
+        let nested_subsys = subsys.start("subsys2", subsys2);
+        sleep(Duration::from_millis(200)).await;
+        subsys
+            .perform_partial_shutdown(nested_subsys)
+            .await
+            .unwrap();
+        set_subsys1_shutdown_performed();
+        subsys.on_shutdown_requested().await;
+        set_subsys1_finished();
+        Ok(())
+    };
+
+    let toplevel = Toplevel::new();
+    let shutdown_token = toplevel.get_shutdown_token().clone();
+
+    tokio::join!(
+        async {
+            let result = toplevel
+                .start("subsys", subsys1)
+                .handle_shutdown_requests(Duration::from_millis(500))
+                .await;
+            assert!(result.is_ok());
+        },
+        async {
+            sleep(Duration::from_millis(300)).await;
+            assert!(subsys1_started.get());
+            assert!(subsys2_started.get());
+            assert!(subsys3_started.get());
+            assert!(!subsys1_finished.get());
+            assert!(subsys2_finished.get());
+            assert!(subsys3_finished.get());
+            assert!(subsys1_shutdown_performed.get());
+            shutdown_token.shutdown();
+        }
+    );
+}
+
+#[tokio::test]
+async fn partial_shutdown_panic_gets_propagated_correctly() {
+    setup();
+
+    let (nested_started, set_nested_started) = Event::create();
+    let (nested_finished, set_nested_finished) = Event::create();
+
+    let nested_subsys = move |subsys: SubsystemHandle| async move {
+        set_nested_started();
+        subsys.on_shutdown_requested().await;
+        set_nested_finished();
+        panic!("Nested panicked.");
+    };
+
+    let subsys1 = move |mut subsys: SubsystemHandle| async move {
+        let handle = subsys.start("nested", nested_subsys);
+        sleep(Duration::from_millis(100)).await;
+        let result = subsys.perform_partial_shutdown(handle).await;
+
+        assert_eq!(result.err(), Some(PartialShutdownError::SubsystemFailed));
+        assert!(nested_started.get());
+        assert!(nested_finished.get());
+
+        subsys.request_shutdown();
+        Ok(())
+    };
+
+    let result = Toplevel::new()
+        .start("subsys", subsys1)
+        .handle_shutdown_requests(Duration::from_millis(500))
+        .await;
+
+    assert!(result.is_ok());
+}
+
+#[tokio::test]
+async fn partial_shutdown_error_gets_propagated_correctly() {
+    setup();
+
+    let (nested_started, set_nested_started) = Event::create();
+    let (nested_finished, set_nested_finished) = Event::create();
+
+    let nested_subsys = move |subsys: SubsystemHandle| async move {
+        set_nested_started();
+        subsys.on_shutdown_requested().await;
+        set_nested_finished();
+        Err(anyhow!("nested failed."))
+    };
+
+    let subsys1 = move |mut subsys: SubsystemHandle| async move {
+        let handle = subsys.start("nested", nested_subsys);
+        sleep(Duration::from_millis(100)).await;
+        let result = subsys.perform_partial_shutdown(handle).await;
+
+        assert_eq!(result.err(), Some(PartialShutdownError::SubsystemFailed));
+        assert!(nested_started.get());
+        assert!(nested_finished.get());
+
+        subsys.request_shutdown();
+        Ok(())
+    };
+
+    let result = Toplevel::new()
+        .start("subsys", subsys1)
+        .handle_shutdown_requests(Duration::from_millis(500))
+        .await;
+
+    assert!(result.is_ok());
+}
+
+#[tokio::test]
+async fn partial_shutdown_during_program_shutdown_causes_error() {
+    setup();
+
+    let (nested_started, set_nested_started) = Event::create();
+    let (nested_finished, set_nested_finished) = Event::create();
+
+    let nested_subsys = move |subsys: SubsystemHandle| async move {
+        set_nested_started();
+        subsys.on_shutdown_requested().await;
+        set_nested_finished();
+        Ok(())
+    };
+
+    let subsys1 = move |mut subsys: SubsystemHandle| async move {
+        let handle = subsys.start("nested", nested_subsys);
+        sleep(Duration::from_millis(100)).await;
+
+        subsys.request_shutdown();
+        sleep(Duration::from_millis(100)).await;
+        let result = subsys.perform_partial_shutdown(handle).await;
+
+        assert_eq!(
+            result.err(),
+            Some(PartialShutdownError::AlreadyShuttingDown)
+        );
+
+        sleep(Duration::from_millis(100)).await;
+
+        assert!(nested_started.get());
+        assert!(nested_finished.get());
+
+        Ok(())
+    };
+
+    let result = Toplevel::new()
+        .start("subsys", subsys1)
+        .handle_shutdown_requests(Duration::from_millis(500))
+        .await;
+
+    assert!(result.is_ok());
+}
+
+#[tokio::test]
+async fn partial_shutdown_on_wrong_parent_causes_error() {
+    setup();
+
+    let (nested_started, set_nested_started) = Event::create();
+    let (nested_finished, set_nested_finished) = Event::create();
+
+    let nested_subsys = move |subsys: SubsystemHandle| async move {
+        set_nested_started();
+        subsys.on_shutdown_requested().await;
+        set_nested_finished();
+        Ok(())
+    };
+
+    let subsys1 = move |mut subsys: SubsystemHandle| async move {
+        let handle = subsys.start("nested", nested_subsys);
+
+        sleep(Duration::from_millis(100)).await;
+
+        let wrong_parent = |child_subsys: SubsystemHandle| async move {
+            sleep(Duration::from_millis(100)).await;
+            let result = child_subsys.perform_partial_shutdown(handle).await;
+            assert_eq!(result.err(), Some(PartialShutdownError::SubsystemNotFound));
+
+            child_subsys.request_shutdown();
+            sleep(Duration::from_millis(100)).await;
+
+            assert!(nested_started.get());
+            assert!(nested_finished.get());
+
+            Ok(())
+        };
+
+        subsys.start("wrong_parent", wrong_parent);
+        subsys.on_shutdown_requested().await;
+
+        Ok(())
+    };
+
+    let result = Toplevel::new()
+        .start("subsys", subsys1)
+        .handle_shutdown_requests(Duration::from_millis(500))
+        .await;
+
+    assert!(result.is_ok());
+}
+
+#[cfg(unix)]
+#[tokio::test]
+async fn shutdown_through_signal() {
+    use nix::sys::signal::{self, Signal};
+    use nix::unistd::Pid;
+
+    setup();
+
+    let subsystem = |subsys: SubsystemHandle| async move {
+        subsys.on_shutdown_requested().await;
+        sleep(Duration::from_millis(200)).await;
+        Ok(())
+    };
+
+    let toplevel = Toplevel::new().catch_signals();
+    tokio::join!(
+        async {
+            sleep(Duration::from_millis(100)).await;
+
+            // Send SIGINT to ourselves.
+            signal::kill(Pid::this(), Signal::SIGINT).unwrap();
+        },
+        async {
+            let result = toplevel
+                .start("subsys", subsystem)
+                .handle_shutdown_requests(Duration::from_millis(400))
+                .await;
+            assert!(result.is_ok());
+        },
+    );
+}
